//
// Created by martin on 19.11.21.
//

#ifndef ANIMTESTBED_SYNCTRACK_H
#define ANIMTESTBED_SYNCTRACK_H

#include <cassert>

#include "3rdparty/json/json.hpp"

constexpr int cSyncTrackMaxIntervals = 8;

/// Metadata used for synced animation blending.
//
// A SyncTrack consists of multiple SyncInterval that are adjacent to each
// other.
//
// Important definitions:
//
// - Absolute Time: time within an animation duration in seconds.
// - Ratio: time relative to the animations duration, e.g. 0.5 corresponds to
//   50% of the duration.
// - SyncTime is a floating point value where the integer parts defines the
//   SyncInterval and the fractional part the fraction within the interval. I.e.
//   a SyncTime of 5.332 means it is ~33% through interval 5.
//
// A SyncInterval is defined by a ratio of the starting point and the ratio of
// the interval's duration.
struct SyncTrack {
  SyncTrack() : m_duration(0.f), m_num_intervals(1) {
    for (int i = 0; i < cSyncTrackMaxIntervals; i++) {
      m_interval_start_ratio[i] = 0.f;
      m_interval_duration_ratio[i] = 0.f;
    }

    m_interval_duration_ratio[0] = 1.0f;
  }

  float m_duration;
  int m_num_intervals;
  float m_interval_start_ratio
      [cSyncTrackMaxIntervals];  //< Starting time of interval in absolute time.
  float m_interval_duration_ratio[cSyncTrackMaxIntervals];  //<

  float CalcSyncFromAbsTime(float abs_time) {
    for (int i = 0; i < m_num_intervals; i++) {
      float query_abs_time = abs_time;
      float interval_start = m_interval_start_ratio[i] * m_duration;
      float interval_end =
          interval_start + m_interval_duration_ratio[i] * m_duration;

      if (query_abs_time < interval_start) {
        query_abs_time += m_duration;
      }
      if (query_abs_time >= interval_start && query_abs_time < interval_end) {
        return float(i)
               + (query_abs_time - interval_start)
                     / (interval_end - interval_start);
      }
    }

    assert(false && "Invalid absolute time");
    return -1.f;
  }

  float CalcRatioFromSyncTime(float sync_time) {
    float interval_ratio = fmodf(sync_time, 1.0f);
    int interval = int(sync_time - interval_ratio);

    return fmodf(
        m_interval_start_ratio[interval]
            + m_interval_duration_ratio[interval] * interval_ratio,
        1.0f);
  }

  bool operator==(const SyncTrack& other) const {
    bool result = m_duration == other.m_duration
                  && m_num_intervals == other.m_num_intervals;

    if (!result) {
      return false;
    }

    for (int i = 0; i < m_num_intervals; i++) {
      if ((fabsf(m_interval_start_ratio[i] - other.m_interval_start_ratio[i])
           > 1.0e-5)
          || (fabsf(
                  m_interval_duration_ratio[i]
                  - other.m_interval_duration_ratio[i])
              > 1.0e-5)) {
        return false;
      }
    }

    return true;
  }

  /** Constructs SyncTrack from markers.
   *
   * Markers are specified in absolute time and must be >= 0 and <= duration.
   * They define the start of the interval. The last marker is implicitly the
   * (possibly looped) first marker.
   */
  static SyncTrack CreateFromMarkers(
      float duration,
      const std::vector<float>& markers) {
    assert(markers.size() > 0);
    assert(markers.size() < cSyncTrackMaxIntervals);

    SyncTrack result;
    result.m_duration = duration;
    result.m_num_intervals = markers.size();

    for (int i = 0; i < markers.size(); i++) {
      assert(markers[i] >= 0.f && markers[i] <= duration);

      int end_index = i == (markers.size() - 1) ? 0 : i + 1;

      float interval_start = markers[i];
      float interval_end = markers[end_index];

      if (interval_end == interval_start) {
        interval_end = interval_start + duration;
      } else if (interval_end < interval_start) {
        interval_end += duration;
      }

      result.m_interval_start_ratio[i] = interval_start / duration;
      result.m_interval_duration_ratio[i] =
          (interval_end - interval_start) / duration;
    }

    return result;
  }

  static SyncTrack
  Blend(float weight, const SyncTrack& track_A, const SyncTrack& track_B) {
    assert(track_A.m_num_intervals == track_B.m_num_intervals);

    SyncTrack result;
    result.m_num_intervals = track_A.m_num_intervals;

    result.m_duration =
        (1.0f - weight) * track_A.m_duration + weight * track_B.m_duration;

    float interval_0_offset =
        track_B.m_interval_start_ratio[0] - track_A.m_interval_start_ratio[0];
    if (interval_0_offset > 0.5f) {
      interval_0_offset = -fmodf(1.f - interval_0_offset, 1.0f);
    } else if (interval_0_offset < -0.5) {
      interval_0_offset = fmodf(1.f + interval_0_offset, 1.0f);
    }

    result.m_interval_start_ratio[0] = fmodf(
        1.0 + (1.0f - weight) * track_A.m_interval_start_ratio[0]
            + weight * (track_A.m_interval_start_ratio[0] + interval_0_offset),
        1.0f);

    for (int i = 0; i < result.m_num_intervals; i++) {
      float interval_duration_A = track_A.m_interval_duration_ratio[i];
      float interval_duration_B = track_B.m_interval_duration_ratio[i];
      result.m_interval_duration_ratio[i] =
          (1.0f - weight) * interval_duration_A + weight * interval_duration_B;

      if (i < cSyncTrackMaxIntervals) {
        result.m_interval_start_ratio[i + 1] =
            result.m_interval_start_ratio[i]
            + result.m_interval_duration_ratio[i];
        if (result.m_interval_start_ratio[i + 1] > 1.0f) {
          result.m_interval_start_ratio[i + 1] =
              fmodf(result.m_interval_start_ratio[i + 1], 1.0f);
        }
      }
    }

    assert(result.m_num_intervals < cSyncTrackMaxIntervals);

    return result;
  }
};

inline void to_json(nlohmann::json& j, const SyncTrack& sync_track) {
  j["type"] = "SyncTrack";
  j["duration"] = sync_track.m_duration;
  for (int i = 0; i < sync_track.m_num_intervals; i++) {
    j["interval"][i]["start"] = sync_track.m_interval_start_ratio[i];
    j["interval"][i]["ratio"] = sync_track.m_interval_duration_ratio[i];
  }
}

inline void from_json(const nlohmann::json& j, SyncTrack& sync_track) {
  assert(j["type"] == "SyncTrack");

  sync_track.m_duration = j["duration"];

  sync_track.m_num_intervals = j["interval"].size();
  assert(sync_track.m_num_intervals < cSyncTrackMaxIntervals);

  for (int i = 0; i < sync_track.m_num_intervals; i++) {
    sync_track.m_interval_start_ratio[i] = j["interval"][i]["start"];
    sync_track.m_interval_duration_ratio[i] = j["interval"][i]["ratio"];
  }
}

#endif  //ANIMTESTBED_SYNCTRACK_H
